#![allow(unused_variables)]
#![allow(unused_assignments)]
#![allow(dead_code)]
#![allow(irrefutable_let_patterns)]

use std::collections::{HashMap, HashSet, VecDeque};
use std::fs::File;
use std::io::{Error, ErrorKind, Read};
use std::ops::{Add, Sub};
use std::sync::mpsc;
use std::sync::mpsc::{Receiver, Sender};
use std::thread;
use rand::Rng;

mod geo_module_as_separate_file;
use geo_module_as_separate_file::distance1;
use geo_module_as_separate_file::calculations::distance as distance_calc; // aliasing
// collision may happen when there are multiple instances of the same name --> like when they are common name or keyword

fn main() {
    crates_basics();
}

fn crates_basics() {
    let mut rng = rand::thread_rng();
    let random_number: i32 = rng.gen();
    println!("{}", random_number);
}

fn modules_basics() {
    let kcle_lat_deg: f64 = 41.4075;
    let kcle_lon_deg: f64 = -81.851111;
    let kslc_lat_deg: f64 = 40.7861;
    let kslc_lon_deg: f64 = -111.9822;

    // let distance = geo_within_same_file::distance(kcle_lat_deg, kcle_lon_deg, kslc_lat_deg, kslc_lon_deg);
    let distance = geo_module_as_separate_file::distance(kcle_lat_deg, kcle_lon_deg, kslc_lat_deg, kslc_lon_deg);
    let distance = distance1(kcle_lat_deg, kcle_lon_deg, kslc_lat_deg, kslc_lon_deg);
    print!("Distance bw the 2 points is {:.1}", distance); // 1 decimal place
}
mod geo_within_same_file {
    const EARTH_RADIUS_IN_KM: f64 = 6371.0;
    pub fn distance(start_lat: f64, start_lon: f64, end_lat: f64, end_lon: f64) -> f64 {
        let kcle_lat_rad = start_lat.to_radians();
        let kslc_lat_rad = end_lat.to_radians();

        let delta_lat = (start_lat - end_lat).to_radians();
        let delta_lon = (start_lon - end_lon).to_radians();

        let inner_circle_angle = f64::powi((delta_lat/2.0).sin(), 2)
            + kcle_lat_rad.cos()
            * kslc_lat_rad.cos()
            * f64::powi(delta_lon/2.0, 2);
        let central_angle = 2.0 * inner_circle_angle.sqrt().asin();

        EARTH_RADIUS_IN_KM * central_angle
    }
}

fn thread_basics() {
    let outer_scope = 382;

    /*  create a new thread
        closure function can use variables in their scope, so the below closure function can use variable outer_scope
        the thread that spawned the below thread keeps going too, but the below spawned thread does not own that data
        hence when outer_scope goes out of scope, the data that below thread is pointing to is also gone
        The variable on the thread has outlived the owner
        So the thread has to take ownership of data to work properly --> we use "move" keyword for the same
        Since now the below thread owns outer_scope, we can't use it in our function that spawned the thread anymore  */
    let join_handle = thread::spawn( move || {
        outer_scope * 2
    });
    // This will stop the code from continuing until the thread is done processing --> blocking
    let result = join_handle.join();
    match result {
        Ok(value) => println!("{}", value),
        Err(error) => println!("{:?}", error)
    }

    let (john_tx, john_rx) = mpsc::channel();
    let (sarah_tx, sarah_rx) = mpsc::channel();

    let john_handle = thread:: spawn(move || {
        john_chat(sarah_tx, john_rx);
    });
    let sarah_handle = thread:: spawn(move || {
        sarah_chat(john_tx, sarah_rx);
    });
}
fn sarah_chat(john_tx: Sender<&str>, sarah_rx: Receiver<&str>) {
    let result = sarah_rx.recv();
    // recv is blocking but try_recv is not, it will see if there is any msg or not and then continue processing
    match result {
        Ok(value) => println!("{}", value),
        Err(_) => {}
    }
    println!("{}", result.unwrap()); // in this case, if error, it will throw a panic
    let _send_result = john_tx.send("Hello John from Sarah!"); //TODO why _ at start of variable
}
fn john_chat(sarah_tx: Sender<&str>, john_rx: Receiver<&str>) {
    let send_result = sarah_tx.send("Hello Sarah from John!");
    let result = john_rx.recv();
    // recv is blocking but try_recv is not, it will see if there is any msg or not and then continue processing
    match result {
        Ok(value) => println!("{}", value),
        Err(_) => {}
    }
    println!("{}", result.unwrap()); // in this case, if error, it will throw a panic
}

fn constraints_basics() {
    let sum = add_sub_2(256, 262);
    println!("{}", sum);
}
fn add<T: Add<Output = T>>(op1: T, op2: T) -> T {
    op1 + op2
}
fn add_sub<T: Add<Output = T> + Sub<Output = T>>(op1: T, op2: T) -> T {
    op1 + op2
}
fn add_sub_2<T>(op1: T, op2: T) -> T
where T: Add<Output = T> + Sub<Output = T>
{
    op1 + op2
}

#[derive(Debug)] // what does this mean
struct NavAid<T, U> {
    name: String,
    freq: T,
    data: U
}
fn generic_types_basics() {

    // lifetime and generics together
    // fn lifetime<`a>(p1: &`a i32) {}
    // fn generic<T>(p1: T) {}
    // fn both<`a, T>(p1: T, p2: &`a i32) {}

    let vor = NavAid {
        name: String::from("DQN"),
        freq: 114.5,
        data: String::from("DQN is a VOR")
    };
    let ndb_data: Option<String> = Option::None;
    let ndb = NavAid {
        name: String::from("HKF"),
        freq: 239,
        data: ndb_data
    };
    println!("VOR info: {:?}", vor);
    println!("NDB info: {:?}", ndb);
}

fn sets_basics() {
    let mut flights = HashSet::new();
    flights.insert(("A123", "11.22"));
    flights.insert(("C123", "6.22"));
    flights.insert(("B123", "20.22"));
    for flight in flights.iter() {
        println!("{:?}", flight);
    }
}

fn maps_basics() {
    let mut flights = HashMap::new();
    flights.insert("B123", ("11.22", "Orlando"));
    flights.insert("A123", ("19.22", "Salt lake"));
    flights.insert("C123", ("1.22", "Berlin"));
    flights.insert("D123", ("13.22", "Boston"));

    let option = flights.get("C123");
    let (time, destination) = option.unwrap();
    println!("{}, {}", time, destination);
    // or
    if let Some((time, destination)) = flights.get("C123") {
        println!("{}, {}", time, destination);
    }

    if !flights.contains_key("S123") {
        flights.insert("S123", ("3.56", "Hawaii"));
    } else {
        println!("Flight already entered!");
    }
}

fn vector_basics() {
    let mut vector1 = vec![1, 2 , 3];
    let mut flights: Vec<&str> = Vec::new();
    flights.push("CLE");
    flights.push("SLC");


    // iter() vs iter_mut()
    for flight in flights.iter() {
        println!("{}", flight);
    }
    flights.insert(1, "MLB");
    flights.remove(2);
    for flight in flights.iter() {
        println!("{}", flight);
    }
    // below will panic in given situation as index 2 no more exists
    // let third = flights[2]; // not preferred due array out of bounds possibility

    let fourth = flights.get(3); // safer as it returns option
    match fourth {
        Some(flight) => println!("{}", flight),
        _ => println!("Not found in array")
    }
    if let Some(flight_value) = flights.get(4) {
        println!("{}", flight_value);
    }
    if let None = flights.get(4) {
        println!("None");
    }


    let mut flightsDeque = VecDeque::new();
    flightsDeque.push_back("CLE");
    flightsDeque.push_front("SLC");
    let count = flightsDeque.len();
    let present = flightsDeque.contains(&"CLE");
    flightsDeque.clear();
}

struct Boeing {
    required_crew: u8,
    range: u16
}
struct Airbus {
    required_crew: u8,
    range: u16
}
trait Flight {
    fn is_legal(&self, required_crew: u8, available_crew: u8, range: u16, distance: u16) -> bool;
}
impl Flight for Boeing {
    fn is_legal(&self, required_crew: u8, available_crew: u8, range: u16, distance: u16) -> bool {
        if available_crew >=required_crew && range + 150 > distance {
            true
        } else {
            false
        }
    }
}
impl Flight for Airbus {
    fn is_legal(&self, required_crew: u8, available_crew: u8, range: u16, distance: u16) -> bool {
        if available_crew >=required_crew && range + 280 > distance {
            true
        } else {
            false
        }
    }
}
fn traits_basics() {
    let boeing = Boeing {
        required_crew: 4,
        range: 7400
    };
    let airbus = Airbus {
        required_crew: 7,
        range: 5400
    };
    let boeing_is_legal = boeing.is_legal(boeing.required_crew, 18, boeing.range, 2358);
    let airbus_is_legal = airbus.is_legal(airbus.required_crew, 3, airbus.range, 2358);
    println!("Boeing legal? {}\nAirbus legal? {}", boeing_is_legal, airbus_is_legal);
}

struct Waypoint {
    name: String,
    latitude: f32,
    longitude: f32
}
struct Segment {
    start: Waypoint,
    end: Waypoint
}
impl Segment { // somehow it has something to do with the above struct definition of Segment
    fn new(start: Waypoint, end: Waypoint) -> Self {
    // here instead of writing Self we can also write Segment(variable/ custom data structure name),
    // but writing Self is more convenient and better for future refactoring like, renaming
        Self {
            start,
            end
        }
    }

    fn distance(&self) -> f32 {
        const EARTH_RADIUS_IN_KM: f64 = 6371.0;
        let start_radians = self.start.latitude.to_radians();
        let end_radians = self.end.latitude.to_radians();
        let delta_latitude = (self.start.latitude - self.end.latitude).to_radians();
        let delta_longitude = (self.start.longitude - self.end.longitude).to_radians();
        let inner_central_angle = f32::powi((delta_latitude / 2.0).sin(), 2)
            + start_radians.cos() * end_radians.cos() * f32::powi((delta_longitude / 2.0).sin(), 2);
        let central_angle = 2.0 * inner_central_angle.sqrt().asin();
        let distance = EARTH_RADIUS_IN_KM as f32 * central_angle;
        distance
    }
}
fn struct_basics() {
    let kcle = Waypoint{
        name: "Cleveland".to_string(),
        latitude: 45.0,
        longitude: 35.0
    };
    let kslc = Waypoint {
        ..kcle
    };
    let kmlb = Waypoint{
        longitude: 78.0,
        ..kslc
    };
}
fn associated_methods() {
    let kcle = Waypoint {
        name: "KCLE".to_string(),
        latitude: 41.4075,
        longitude: -81.851111
    };
    let kslc = Waypoint {
        name: "KSLC".to_string(),
        latitude: 40.7861,
        longitude: -111.9822
    };

    let kcle_kslc = Segment::new(kcle, kslc);
    println!("{:.1}", kcle_kslc.distance());
}

fn error_propagating_file_base() {
    let filename = "C:\\Users\\rarani\\temp2.txt";
    let file_data = error_propagating_file(filename);
    // println!("{:#?}", file_data);
    match file_data {
        Ok(data) => println!("{}", data),
        Err(_) => println!("Error encountered!")
    }
}
fn error_propagating_file(filename: &str) -> Result<String, Error> {
    let mut file_handle = File::open(filename)?;
    let mut file_data = String::new();
    file_handle.read_to_string(&mut file_data)?;
    Ok(file_data)
}

fn error_handling_file() {
    let filename = "C:\\Users\\rarani\\temp.txt";
    let file_handle = File::open(filename);
    match file_handle {
        Ok(file) => {
            println!("{:#?}", file);
        },
        Err(error) => {
            println!("{:#?}", error);
            match error.kind() {
                ErrorKind::NotFound => {
                    match File::create(filename) {
                        Ok(file) => {
                            println!("File created");
                        },
                        Err(error) => {
                            println!("{:#?}", error);
                        }
                    }
                },
                _ => {
                    println!("{:#?}", error);
                }
            }
        }
    }
}

fn error_handling() {
    let vector = vec![1, 3, 5, 6];
    println!("{}", vector[9]);
}

fn closure_basics() {
    let name = "Duck Airlines";
    let write_message = |slogan| -> String {
        String::from(format!("{}. {}", name, slogan))
    };
    let phrase = write_message(String::from("We hit the ground everytime!"));
    println!("{}", phrase);
}

fn ownership_borrowing_with_functions() {
    let mut original = String::from("original value");
    println!("Initial original: {}", original);
    print_original(&original);
    change_original(&mut original);
    println!("Final value{}", original);
}
fn print_original(original: &String) {
    println!("fn print_original: {}", original);
}
fn change_original(original: &mut String) {
    let next = original; // let mut next = original ---> mut is not required
    *next = String::from("next value");
    println!("fn change_original: {}", next);
}

fn return_greater(first: u8, second: u8) -> u8{
    if first > second {
        first
    } else {
        second
    }
}

fn lifetime_basics() {
    let referenced_int = 8;
    let returned_value =return_one_parameter(&referenced_int);
    println!("{}", returned_value);
    let val1 = 23;
    let val2 = 234;
    let val = explicit_lifetime(&val1, & val2);
    println!("{}", val);
}
fn return_one_parameter(value: &i32) -> &i32 {
    value
}
fn explicit_lifetime<'a>(p1: &'a i32, p2: &'a i32) -> &'a i32 {
    if p1 > p2 {
        p1
    } else {
        p2
    }
}

fn borrowing_basics4() {
    let mut original = String::from("original value");
    println!("{}", original);
    {
        let next = &mut original; // so that next is not just read only
        *next = String::from("next value"); // *next --> de-referencing the variable
        println!("{}", next);
        println!("{}", original);
    }
    original = String::from("new value");
    println!("{}", original);
}

fn borrowing_basics3() {
    let mut original = String::from("original value");
    println!("{}", original);

    let next = &original;
    original = String::from("new value");
    // While there are read only references to a variable being used downstream, we cannot change the data of that variable
    // error thrown at below line because original has been borrowed by next and we cannot assign a value to a borrowed variable
    // println!("{}", next);
}

fn borrowing_basics2() {
    let mut original = String::from("original value");
    println!("{}", original);
    original = String::from("new value");

    let next = &original;
    println!("{}", next);
}

fn borrowing_basics1() {
    let original = String::from("original value");
    println!("{}", original);
    let next = &original; // next has read only access to data in original
    println!("{}", next);
}

fn not_borrowing_basics() {
    let original = String::from("original value");
    println!("{}", original);
    let next = original;
    // println!("{}", original); // error
}

fn distance_bw_waypoints() {
    const EARTH_RADIUS_IN_KM: f64 = 6371.0;
    let route = [
        ("KCLE", 41.4075, -81.851111),
        ("BHDEF", 40.4075, -80.851111),
        ("BKUJWAH", 42.4075, -101.851111),
        ("JUSDH", 45.4075, -91.851111),
        ("KCLE", 44.4075, -111.851111)
    ];

    let mut total_distance = 0.0;
    let mut prev_waypoint: Option<(&str, f64, f64)> = None;

    for waypoint in route {
        match prev_waypoint {
            None => {
                prev_waypoint = Option::from(waypoint.clone());
                continue;
            },
            Some(prev_waypoint_value) => {
                const EARTH_RADIUS_IN_KM: f64 = 6371.0;

                let prev_lat_rad = prev_waypoint_value.1.to_radians();
                let curr_lat_rad = waypoint.1.to_radians();

                let delta_lat = (prev_waypoint_value.1 - waypoint.1).to_radians();
                let delta_lon = (prev_waypoint_value.2 - waypoint.2).to_radians();

                let inner_circle_angle = f64::powi((delta_lat/2.0).sin(), 2)
                    + prev_lat_rad.cos()
                    * curr_lat_rad.cos()
                    * f64::powi(delta_lon/2.0, 2);
                let central_angle = 2.0 * inner_circle_angle.sqrt().asin();
                let distance = EARTH_RADIUS_IN_KM * central_angle;

                total_distance += distance;
                prev_waypoint = Option::from(waypoint.clone());

                println!("Distance bw {} and {} is {:.1}", prev_waypoint_value.0, waypoint.0, distance);
            }
        }
    }
}

fn for_loop_basics() {
    for index in 1..=10 {
        println!("{}", index);
    }
    let duck_aircraft = ["a 1", "b 2", "c 3"];
    for aircraft in duck_aircraft.iter() {
        println!("{}", aircraft);
    }
    for aircraft in duck_aircraft {
        println!("{}", aircraft);
    }
}

fn while_loop_and_while_let_loop() {
    let mut counter = 0;
    while counter <= 10 {
        println!("{}", counter);
        counter += 1;
    }
}

fn rust_loop() {
    let mut counter = 0;
    loop {
        counter += 1;
        println!("{}", counter);
        if counter == 5 {
            continue;
        }
        if counter == 10 {
            break;
        }
    }
}

fn if_let_basics() {
    let animal = "Duck";

    if let animal = "Duck" {
        print!("{}", animal); // Duck
    }
    if let animal = "Dog" {
        print!("{}", animal); // Dog
    }
}

enum NavAids {
    NDB(u16),
    VOR(String, f32),
    VORDME(String, f32),
    FIX {name: String, lat: f32, lon: f32} // tuple
}
fn print_nav_aids(nav_aid: &NavAids) {
    match nav_aid {
        NavAids::NDB(khz) => print!("NDB frequency is {} kilohertz", khz),
        NavAids::VOR(id, freq) => print!("VOR id is {} and frequency is {} kilohertz", id, freq),
        NavAids::VORDME(id, freq) => print!("VORDME id is {} and frequency is {} kilohertz", id, freq),
        NavAids::FIX {name, lat, lon} => print!("FIX {} is at {} lat and {} lon", name, lat, lon),

    }
}
fn match_enum_basics() {
    let ndb_uwl = NavAids::NDB(385);
    let vor_dqn = NavAids::VOR(String::from("DQN"), 114.5);
    let vor_dme_sgh = NavAids::VORDME(String::from("SGH"), 113.2);
    let fix_tarry = NavAids::FIX{
        name: String::from("Tarry"),
        lat: 40.05333,
        lon: -83.91367
    };
    print_nav_aids(&vor_dme_sgh);
}

fn match_basics() {
    let animal = "Duck";
    match animal {
        "Duck" => print!("Quack!"),
        "Dog" => print!("Bark!"),
        _ => println!("All quiet out here")
    }

    let ndb_freq:u16 = 200;
    match ndb_freq {
        200..=500 => { // inclusive range
            print!("Valid frequency");
        }
        _ => {
            print!("Invalid frequency");
        }
    }

    match ndb_freq {
        ndb_freq if ndb_freq >= 200 && ndb_freq <= 500 => {
            print!("Valid");
        }
        _ => {
            print!("Invalid");
        }
    }
}

fn option_basics() {
    let phrase = String::from("Duck Airlines");
    let letter = phrase.chars().nth(5);
    // let letter = phrase.chars().nth(15);
    let value = match letter {
        Some(ch) => ch.to_string(),
        None => String::from("No value")
    };
    println!("{}", value)
}

fn enum_basics() {
    enum NavAids1 {
        NDB, // 0
        VOR, // 1
        VORDME // 2
    }
    enum NavAids2 {
        NDB = 5, // 5
        VOR, // 6
        VORDME // 7
    }
    enum NavAids3 {
        NDB, // 0
        VOR = 8, // 8
        VORDME // 9
    }
    enum NavAids4 {
        NDB = 16, // 16
        VOR = 8, // 8
        VORDME = 76 // 76
    }
    enum NavAids5 {
        NDB,
        VOR,
        VORDME,
        FIX {name: String, lat: f32, lon: f32} // tuple
    }
    enum NavAids6 {
        NDB(u16),
        VOR(String, f32),
        VORDME(String, f32),
        FIX {name: String, lat: f32, lon: f32} // tuple
    }
    print!("NDB:\t{}", NavAids1::NDB as u8);
    print!("VOR:\t{}", NavAids1::VOR as u8);
    print!("VORDME:\t{}", NavAids1::VORDME as u8);
}

fn distance_bw_slc_and_cle() {
    const EARTH_RADIUS_IN_KM: f64 = 6371.0;
    let kcle_lat_deg: f64 = 41.4075;
    let kcle_lon_deg: f64 = -81.851111;
    let kslc_lat_deg: f64 = 40.7861;
    let kslc_lon_deg: f64 = -111.9822;

    let kcle_lat_rad = kcle_lat_deg.to_radians();
    let kslc_lat_rad = kslc_lat_deg.to_radians();

    let delta_lat = (kcle_lat_deg - kslc_lat_deg).to_radians();
    let delta_lon = (kcle_lon_deg - kslc_lon_deg).to_radians();

    let inner_circle_angle = f64::powi((delta_lat/2.0).sin(), 2)
        + kcle_lat_rad.cos()
        * kslc_lat_rad.cos()
        * f64::powi(delta_lon/2.0, 2);
    let central_angle = 2.0 * inner_circle_angle.sqrt().asin();

    let distance = EARTH_RADIUS_IN_KM * central_angle;
    // TODO here the output is 2542.6 but in the video, the output 2513.0 (https://app.pluralsight.com/course-player?clipId=2e7330fb-2c64-4ebd-a061-41ea786ec7f6)
    print!("Distance bw the 2 points is {:.1}", distance); // 1 decimal place
}

fn basics1() {
    let x:u32;
    // let location: [f32;2] = [41.409469, -81.8546911]; //array
    // let location = [0.0; 1000];
    let location= ("KCLE", 41.409469, -81.8546911); // tuple
    // println!("Location name: {}, latitude: {}, longitude: {}", location.0, location.1, location.2);
    let (name, latitude, longitude) = location;
    println!("Location name: {}, latitude: {}, longitude: {}", name, latitude, longitude);

    let person_name_slice = "Raj Rani";
    let person_name_string1 = person_name_slice.to_string();
    let person_name_string2 = "Raj Rani".to_string();
    let person_name_string3 = String::from("Raj Rani");

    //here string data isn't copied to the string slice
    let person_name_slice2 = &person_name_string1;
    //de-referencing the variable
    let person_name_slice3 = person_name_string1.as_str();

    let duck = "Duck";
    let airlines = "Airlines";

    let airline_name = [duck, " ", airlines].concat();
    let airline_name2 = format!("{} {}", duck, airlines);
    println!("{}", airline_name2);

    let mut slogan = String::new();
    slogan.push_str("We hit the ground");
    slogan.push(' ');
    slogan = slogan + "every time";
    println!("{}", slogan);

    let my_inferred_variable = 5_i8;

    let float_thirty_two: f32 = 17.2;
    let unsigned_eight: u8 = 5;
    let cast_unsigned_eight = unsigned_eight as f32;
    let result = float_thirty_two / cast_unsigned_eight;
    println!("{}", result);

    let number: u8 = 65;
    let letter: char = number as char;
    // only u8 can be cast to char as char is 4 bytes long but u128 is 16 bits long
    println!("{}", letter);


    // shadowing
    let scope_test = "string";
    let scope_test = 90;
    println!("{}", scope_test); // prints 90

    let squared = i32::pow(8, 2);
    let float_integer = f32::powi(6.5,2);
    let float_float = f32::powf(6.5,2.5);
}
